#! /usr/bin/env casarun
#
#
#   admit1.py :  a first example ADMIT pipeline/flow
#
#   Usage:      $ADMIT/admit/test/admit1.py  [line.fits [alias]] [cont.fits]
#
#   this will create a line.admit directory with all the
#   BDP's and associated data products inside. Linecubes will
#   have their own directory with their associates products
#   in that directory.
#   Optionally you can give a continuum image, cont.fits
#
#   if you give an admit directory, it will try and re-run in that directory.
#   although it is suggested to use admit0.py as a template and edit this.

version  = '10-sep-2015'

#  ===>>> set some parameters for this run <<<=================================================================
#
#  !!! do not change these defaults, these are meant to be our desired defaults    !!!
#  !!! in an ideal universe. Instead, use the commented section below to enable    !!!
#  !!! them, or add your favorites to some new value to experiment with.           !!!

import admit.util.PlotControl as PlotControl

file     = 'foobar.fits'    # the default FITS input name to be ingested
cont     = ''               # add this continuum fits file to the ingestion process?
useMask  = True             # use a mask where fits data == 0.0
alias    = ''               # use a short alias instead of the possibly long basename?
vlsr     = None             # either set it below, or make get_vlsr() to work (else vlsr=0 will be used)
maxpos   = []               # default to the peak in the cube for CubeSpectrum
robust   = ('hin',1.5)      # default hinges-fences
clean    = True             # clean up the previous admit tree, i.e. no re-running
plotmode = PlotControl.BATCH # NOPLOT, BATCH, INTERACTIVE, SHOW_AT_END
plottype = PlotControl.PNG   # PNG, JPG, etc.
loglevel = 15               # 10=DEBUG, 15=TIMING 20=INFO 30=WARNING 40=ERROR 50=FATAL
insmooth = []               # smooth inside of Ingest_AT, in pixels
smooth   = []               # if set, smooth the cube right after ingest
usePeak  = True             # LineCubeSpectra through peak of a Moment0 map? (else pos[] or SpwCube Peak)
useCSM   = False            # if usePeak, should CubeSum (CSM) be used (instead of mom0 from LineCube)
pvslice  = []               # PV Slice (x0,y0,x1,y1)
pvslit   = []               # PV Slice (xc,yc,len,pa)  if none given, it will try and find out
useUID   = False            # True = LineUID is more robust, but no real lineID  
usePPP   = True             # create and use PeakPointPlot?
useMOM   = True             # if no lines are found, do a MOM0,1,2 anyways ?
pvSmooth = [10,10]          # smooth the PVslice ?   Pos or Pos,Vel pixel numbers

# -- above here are sensible default meant to automate a flow. Don't edit them above this line --
# -- below here you can tinker with some of the above reasonable defaults                      --
#maxpos   = [174,160]           # N253.ce cubes
#maxpos   = get_pos(6)+get_pos(5)     # the RA,DEC positions from table in Meier et al. (use 1..10 here, 0 to disable)
#maxpos   = [170,170,160,160]   # two positions 
#maxpos   = [70,105,100,107]    # test21.fits
#robust  = ('hinge',-1)   # stats method, you need CASA 4.4 or above for this
#robust  = ('hinges-fences',1.5)          # stats method, you need CASA 4.4 or above for this
#robust  = ('fit-half','mean')            # stats method, you need CASA 4.5 or above for this
#robust  = ('fit-half','median', False)       # stats method, you need CASA 4.4 or above for this
#robust  = ('fit-half','zero')           # stats method, you need CASA 4.4 or above for this
#robust  = ('classic','tiled')            # stats method, you need CASA 4.4 or above for this
#robust  = ('chauvenet',-1)               # stats method, you need CASA 4.4 or above for 
robust  = ()
#pvslice = [78.5,227.1,270.4,100.1]    # n253.ce slice (start,end)
#pvslit = [174.0,163.0,250.0,57.0]    # n253.ce slit (center,len,pa)
# pvslit = [158,173,250,57]           # right on pos=#7
#pvslice = [10,10,90,90]    # 
#### N6503:  17h49m26.4s +70d08m40s 
#pvslit = [162,122,360,120.0]       # N6503-1996 VLA data len=415?
#pvslit = [128,128,200,120.0]       # N6503-1983 VLA data len=285?
#pvslit = [248,242,300,57.0]         # sb185 along a disk?
#pvslice = [104,13,141,187]   # Bittle ic10
#pvSmooth = []                 
#insmooth = [4,4]            # image smooth during ingest (gaussian pixels)    BUG: Ingest works around loosing OBJECT name
#smooth = [4,4]              # image smooth right after ingest (Smooth_AT)     BUG: _do_plot ?
#loglevel = 20               # 10=DEBUG, 15=TIMING 20=INFO 30=WARNING 40=ERROR 50=FATAL
#vlsr = 6.9                  # test1 will not detect it at 6.9, but ok at 7.0
#usePeak = False
#useCSM  = True
#useMask  = False
#useMOM = False
#usePPP = False
#useUID = True
#plotmode = PlotControl.INTERACTIVE

# =================================================================================================================
# python system modules
import sys, os, math

# essential admit modules
from admit.AT import AT
import admit.Admit as admit
import admit.util.bdp_types as bt
import admit.util.VLSR as VLSR

# AT's we need (we normally don't need BDP's in user code such as this)
from admit.at.Ingest_AT          import Ingest_AT
from admit.at.Smooth_AT          import Smooth_AT
from admit.at.CubeStats_AT       import CubeStats_AT
from admit.at.CubeSpectrum_AT    import CubeSpectrum_AT
from admit.at.CubeSum_AT         import CubeSum_AT
from admit.at.FeatureList_AT     import FeatureList_AT
from admit.at.LineID_AT          import LineID_AT
from admit.at.LineUID_AT         import LineUID_AT
from admit.at.LineCube_AT        import LineCube_AT
from admit.at.Moment_AT          import Moment_AT
from admit.at.OverlapIntegral_AT import OverlapIntegral_AT
from admit.at.PVSlice_AT         import PVSlice_AT
from admit.at.PVCorr_AT          import PVCorr_AT

def get_pos(i):
    """ return key positions in N253 (1..10) from Meier's Table 2:
         0 = blank, if you want to use the peak in the cube
        11 = map center, the reference position of N253
    """
    pos = [ [],                                     # 0 = blank
            ['00h47m33.041s',	'-25d17m26.61s'	],  # pos 1
            ['00h47m32.290s',	'-25d17m19.10s'	],  #     2
            ['00h47m31.936s',	'-25d17m29.10s'	],  #     3
            ['00h47m32.792s',	'-25d17m21.10s'	],  #     4
            ['00h47m32.969s',	'-25d17m19.50s'	],  #     5
            ['00h47m33.159s',	'-25d17m17.41s'	],  #     6
            ['00h47m33.323s',	'-25d17m15.50s'	],  #     7
            ['00h47m33.647s',	'-25d17m13.10s'	],  #     8
            ['00h47m33.942s',	'-25d17m11.10s'	],  #     9
            ['00h47m34.148s',	'-25d17m12.30s'	],  # pos 10
            ['00h47m33.100s',   '-25d17m17.50s' ],  # map reference 
          ]
    return pos[i]

# 0:47:33.317 -25.17.15.926 = pos 7

# placeholder until we have an official way to find the VLSR of a source
# this is a remnant from how we did this in milestone2.py
# See the file $ADMIT/etc/vlsr.tab if you want to add sources
def get_vlsr(source, vlsr=None):
    if vlsr != None : return vlsr
    vq = VLSR.VLSR()
    return vq.vlsr(source.upper())

def print_linelist(a,lineid):
    """print the linelist, from lineid's BDP_OUT[0], 
    but make sure LineID has been run
    """
    a.run()
    a.write()
    b1 = a[lineid][0]
    cols = b1.table.getHeader()
    n  = 0;
    row = b1.table.getRow(n)
    print "LineList:"
    while row != None:
        ch0   = row[cols.index("startchan")]
        ch1   = row[cols.index("endchan")]
        uid   = row[cols.index("uid")]
        sigma = row[cols.index("sigma")]
        fwhm  = row[cols.index("fwhm")]
        peak  = 1000*row[cols.index("peakintensity")]
        logp  = math.log10(peak)
        print "Line: %d %d  %-30s  %.2f  %.2f    %.2f %.2f" % (ch0,ch1,uid,sigma,fwhm,peak,logp)
        n = n + 1
        row = b1.table.getRow(n)


def admit_dir(file):
    """ create the admit directory name from a filename 
    This filename can be a FITS file (usually with a .fits extension
    or a directory, which would be assumed to be a CASA image or
    a MIRIAD image
    It can be an absolute or relative address
    """
    loc = file.rfind('.')
    ext = '.admit'
    if loc < 0:
        return file + ext
    else:
        if file[loc:] == ext:
            print "Warning: assuming a re-run on existing ",file
            return file
        return file[:loc] + ext

def get_admit_vars(filename):
    print "new"


def get_admit_vars1(module_name):
    module = globals().get(module_name, None)
    book = {}
    if module:
        book = {key: value for key, value in module.__dict__.iteritems() if not (key.startswith('__') or key.startswith('_'))}
    return book

# Before command line parsing, attempt to find 'admit_vars.py' with variables to override the admit vars here
# this doesn't work yet, since CASA modifies the python environment
try:
    print 'Trying admit_vars'
    import admit_vars
    book = get_admit_vars('admit_vars')
    for key,val in book.iteritems():
        # print "ADMIT_VAR: ",key,val,type(key),type(val)
        exec(key + '=' + repr(val))
except:
    print "No admit_vars.py found, and that's ok."


# allow a command line argument to be the fits file name, unless you have foobar.fits ;-)
argv = admit.casa_argv(sys.argv)
if len(argv) > 1:
    file  = argv[1]
    alias = ""
if len(argv) > 2:
    file  = argv[1]
    alias = argv[2]
if len(argv) > 3:
    file  = argv[1]
    alias = argv[2]
    cont  = argv[3]

#-----------------------------------------------------------------------------------------------------------------------
#------------------------------------------------------- start of script -----------------------------------------------

#  announce version
print 'ADMIT1: Version ',version

#  do the work in a proper ".admit" directory
adir = admit_dir(file)
#   dirty method, it really should check if adir is an admit directory
if clean and adir != file:
    print "Removing previous results from ",adir
    os.system('rm -rf %s' % adir)
    create=True
else:
    create=False

a = admit.Admit(adir,name='Testing ADMIT1 style pipeline - version %s' % version,create=create,loglevel=loglevel)

if a.new:
    print "Starting a new ADMIT using ",argv[0]
    os.system('cp -a %s %s' % (argv[0],adir))
    a.set(admit_dir=adir)
else:
    print "All done, we just read an existing admit.xml and it should do nothing"
    print "Use admit0.py to re-run inside of your admit directory"
    #
    a.fm.diagram('admit.dot')
    a.show()
    a.showsetkey()
    sys.exit(0)

# Default ADMIT plotting environment
a.plotparams(plotmode,plottype)

# Ingest
ingest1 = a.addtask(Ingest_AT(file=file,alias=alias))
a[ingest1].setkey('mask',useMask) 
a[ingest1].setkey('smooth',insmooth) 
#a[ingest1].setkey('pbcorr','nosuchfile')                # <-- does not work yet
#a[ingest1].setkey('box',[0,0,10,99,99,20])          # blc trc in 3D
#a[ingest1].setkey('symlink',True)
#a[ingest1].setkey('box',[0,0,10,99,99,20])          # blc trc in 3D
#a[ingest1].setkey('box',[200,200,300,300])          # blc trc in 2D, take all channels?
#a[ingest1].setkey('box',[85,85,255,255])            # blc trc in 2D, take all channels?    spw1 proof for pvslice
#a[ingest1].setkey('box',[10,10,20,20])              # blc trc in 2D, take all channels?    spw1 proof for pvslice
#a[ingest1].setkey('box',[200,200,320,320])          # blc trc in 2D, take all channels?    spw1 proof for 
#a[ingest1].setkey('box',[0,0,13,370,250,77])  # n6503 narrow test
#a[ingest1].setkey('box',[10,10,10,20,20,20])
#a[ingest1].setkey('box',[1,49]) 
#a[ingest1].setkey('box',[10,10,20,20])
#a[ingest1].setkey('edge',[10,10])
#a[ingest1].setkey('box',[10,10,10,350,240,80])
bandcube1 = (ingest1,0)   

if len(smooth) > 0:
    smooth1 = a.addtask(Smooth_AT(), [bandcube1])
    #a[smooth1].setkey('bmaj',7.0) 
    #a[smooth1].setkey('bmin',7.0) 
    #a[smooth1].setkey('bpa',50.0) 
    bandcube2 = (smooth1,0)
    # forget about the original, so we can continue the flow with the smoothed cube
    # in this model, it would be better to use insmooth= for Ingest_AT, it doesn't
    # need the extra space to hold bandcube1
    bandcube1 = bandcube2

#a.exit(1)

if cont != '':
    ingest2 = a.addtask(Ingest_AT(file=cont,alias=alias+'-cont'))
    contmap = (ingest2,0)
    #
    cubestats2 = a.addtask(CubeStats_AT(), [contmap])
    #a[cubestats1].setkey('robust',robust)
    csttab2 = (cubestats2,0)
    featurelist2 = a.addtask(FeatureList_AT(), [contmap])

# CubeStats - will also do log(Noise),log(Peak) plot
cubestats1 = a.addtask(CubeStats_AT(), [bandcube1])
a[cubestats1].setkey('robust',robust)
a[cubestats1].setkey('ppp',usePPP)
csttab1 = (cubestats1,0)


#a.exit(1)

# CubeSum
moment1 = a.addtask(CubeSum_AT(), [bandcube1,csttab1])
a[moment1].setkey('numsigma',4.0)   # Nsigma clip in cube
a[moment1].setkey('sigma',99.0)     # >0 force single cuberms sigma from cubestats; <0 would use rms(freq) table
#a[moment1].setkey('sigma',-1.0)
csmom0 = (moment1,0)

if False:
    test = csmom0      # test 2D
    #test = bandcube1  # test 3D
    oi1 = a.addtask(OverlapIntegral_AT(), [test,test,test])
    #a[oi1].setkey('chans',[10,20])    # test 3D
    a[oi1].setkey('alias','test1')
    oi2 = (oi1,0)

#a.exit(1)


# CubeSpectrum 
if len(maxpos) > 0:
    cubespectrum1 = a.addtask(CubeSpectrum_AT(), [bandcube1])
    a[cubespectrum1].setkey('pos',maxpos)
else:
    #cubespectrum1 = a.addtask(CubeSpectrum_AT(), [bandcube1,csttab1])
    cubespectrum1 = a.addtask(CubeSpectrum_AT(), [bandcube1,csmom0])
csptab1 = (cubespectrum1,0)

#a.exit(1)

# PVSlice
if len(pvslice) == 4:
    # hardcoded with a start,end
    slice1 = a.addtask(PVSlice_AT(slice=pvslice,width=11),[bandcube1])
elif len(pvslit) == 4:
    # hardcoded with a center,len,pa
    slice1 = a.addtask(PVSlice_AT(slit=pvslit,width=11),[bandcube1])
else:
    # use cubesum's map to generate the best slice
    # the PPP method didn't seem to work so well yet, and is slow, so still commented out here
    #slice1 = a.addtask(PVSlice_AT(width=5),[bandcube1,csttab1])     <--  BUG with summary needed a figure CSM style input makes
    slice1 = a.addtask(PVSlice_AT(width=5),[bandcube1,csmom0])
    a[slice1].setkey('clip',0.3)        # TODO: this is an absolute number for mom0
    a[slice1].setkey('gamma',1.0)       # SB185 works better with gamma=4
a[slice1].setkey('smooth',pvSmooth)     # smooth, in pixel numbers
pvslice1 = (slice1,0)

corr1 = a.addtask(PVCorr_AT(),[pvslice1,csttab1])
#a[corr1].setkey('chans',[530,584])     # this is automated, but here you pick a channel range line
#a[corr1].setkey('nchan',30)
pvcorr1 = (corr1,0)

#a.exit(1)

# RUN:: need to run now, since LineID_AT needs the vlsr
a.run()
a.write()
print "OBJECT 1:", a.summaryData.get('object')
vlsr = get_vlsr(a.summaryData.get('object')[0].getValue()[0],vlsr)
print "VLSR = ",vlsr

# lineID  
#lineid1 = a.addtask(LineID_AT(vlsr=vlsr), [csttab1, csptab1])
# method  = 2   # 1=asap   2=astute
#lineid2 = a.addtask(LineID_AT(vlsr=vlsr,method=2), [csttab1])        
#lineid3 = a.addtask(LineUID_AT(vlsr=vlsr,method=1), [csptab1]) 
if useUID:
    # LineUID uses method=<INTEGER>
    # LineUID with a csp usually doesn't make sense
    lineid4 = a.addtask(LineUID_AT(vlsr=vlsr,method=2), [csttab1]) 
    print "####LINEUID TASK #=%d" % lineid4
    #    4/6/3 didn't work for sb179
    a[lineid4].setkey('pmin',4.0)
    a[lineid4].setkey('minchan',6)
    a[lineid4].setkey('maxgap',5)    # 20 for the SB outflows
    a[lineid4].setkey('bottom',True)
    lltab1 = (lineid4,0)
else:
    # LineID uses integer segment=<STRING>:   pick ASAP or ADMIT                  #  --- comments on success or failure on test21.fits ---
                                                                                  #  ok=test21.fits    ok0,1,2,3 = n253_spw0,1,2,3
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ADMIT"), [csptab1])         # ok
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ASAP"), [csptab1])          # ok
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ADMIT"), [csttab1])         # ok
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ASAP"), [csttab1])          # ok
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ASAP"), [csttab1,csptab1])  # ok
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ASAP"), [csptab1,csttab1])  # ok
    # >>> the next list is the recommended default <<<
    lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ADMIT"), [csptab1,csttab1])  #     

    ####  all below here in UID mode
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ASAP"), [csptab1])          # ok
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ADMIT"), [csptab1])         # ok
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ASAP"), [csttab1])          # ok 
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ADMIT"), [csttab1])         # ok
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ASAP"), [csttab1,csptab1])  # ok
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ASAP"), [csptab1,csttab1])  # ok
    #lineid1 = a.addtask(LineID_AT(vlsr=vlsr,segment="ADMIT"), [csptab1,csttab1]) # ok

    #a[lineid1].setkey('donotIdentify',True)
    #a[lineid1].setkey('recombLevel','OFF')
    #a[lineid1].setkey('recombLevel','DEEP')
    #a[lineid1].setkey('pmin',4.0)
    #a[lineid1].setkey('mode','ALL')
    #a[lineid1].setkey('segment','ASAP')         
    #a[lineid1].setkey('smooth',(None,0))
    lltab1 = (lineid1,0)

if False:
    #print_linelist(a,lineid1)
    #print_linelist(a,lineid2)
    print_linelist(a,lineid3)
    print_linelist(a,lineid4)

#a.exit(1)
a.run()
a.write()

nlines = len(a[lltab1[0]][0])
print "nlines=",nlines


# LineCube
linecube1 = a.addtask(LineCube_AT(), [bandcube1,lltab1])
#a[linecube1].setkey('regrid',15.0)
#a[linecube1].setkey('grow',-400)      # -width
#a[linecube1].setkey('grow',20)         # +growth
#a[linecube1].setkey('width',20.0)      # width in km/s [not functional yet???]
#a[linecube1].setkey('regrid',0.1)      # channels in km/s   [not implemented yet???]

#a.exit(1)

# RUN_1: now we need to run the flow, since we need to 
# know the number of Lines found and produce the linecubes 
# for the next for-loop.
a.run()
a.write()


nlines = len(a[linecube1])
print "Found %d lines during runtime" % nlines

x = range(nlines)    # place holder to contain mol/line
m = {}               # task id for moment on this mol/line
sp= {}               # task id for cubespectrum on this mol/line
st= {}               # task id for cubestats

# loop over all lines  
# produce linecubes, moments and spectra
for i in range(nlines):
    x[i] = a[linecube1][i].getimagefile(bt.CASA)
    print "LineDir:", i, x[i]
    # Moment maps from the LineCube
    linecubei = (linecube1,i)
    m[x[i]] = a.addtask(Moment_AT(),[linecubei,csttab1])
    print "MOMENT_AT:",m[x[i]]
    a[m[x[i]]].setkey('moments',[0,1,2])
    #a[m[x[i]]].setkey('cutoff',[2.0,3.0,3.0])
    a[m[x[i]]].setkey('numsigma',[2.0])
    a[m[x[i]]].setkey('mom0clip',2.0)        # TODO: this is still an absolute value
    if usePeak:
        if useCSM:
            momenti0 = csmom0               # CubeSum mom0
        else:
            momenti0 = (m[x[i]],0)          # this linecube mom0
        # CubeSpectrum through the line cube where the mom0 map has peak
        sp[x[i]] = a.addtask(CubeSpectrum_AT(),[linecubei,momenti0])
    elif len(maxpos) > 0:
        # CubeSpectrum through the line cube at given point, 
        # use the same maxpos for all linecubes as we used before
        sp[x[i]] = a.addtask(CubeSpectrum_AT(),[linecubei])
        a[sp[x[i]]].setkey('pos',maxpos)
        # CubeStats
        st[x[i]] = a.addtask(CubeStats_AT(),[linecubei])        
    else:
        # CubeSpectrum last resort, where it takes the peak from cube maximum
        sp[x[i]] = a.addtask(CubeSpectrum_AT(),[linecubei,csttab1])

if useMOM and nlines == 0:
    # if no lines found, take a moment (0,1,2) over the whole cube
    moment1 = a.addtask(Moment_AT(),[bandcube1,csttab1])
    a[moment1].setkey('moments',[0,1,2])
    a[moment1].setkey('numsigma',[2.0])
    a[moment1].setkey('mom0clip',2.0)       # Nsigma

# final run and save
a.run()
a.write()

# check with Marc is that function is really useful?
linelist = a.summaryData.getLinelist()
print "LINES FOUND: " 
for line in linelist:
    print line

# symlink resources and write out index.html
a.updateHTML()

#a.showsetkey()
#print "OBJECT:", a.summaryData.get('object')[0].getValue()[0]
# accumulate and print the dataSummary
#a.print_summary()

